package filegateway

import (
	"context"
	"errors"
	"io"
	"os"
	"path"
	"strings"
	"time"

	"cloud.google.com/go/storage"
	"github.com/rs/zerolog"
	"google.golang.org/api/googleapi"
	"google.golang.org/api/iterator"
	"google.golang.org/api/option"
	xbus "xbus.io/go-xbus/v4"
)

// GCSFilesystem stores files on a Google Cloud Storage
type GCSFilesystem struct {
	FilesystemOptions

	client *storage.Client
	bucket *storage.BucketHandle
}

func gcsFilesystemFactory(options FilesystemOptions, log zerolog.Logger) (Filesystem, error) {
	var clientOptions []option.ClientOption
	if endpoint, err := options.GetCustomStringD("endpoint", ""); endpoint != "" {
		clientOptions = append(clientOptions, option.WithEndpoint(endpoint))
	} else if err != nil {
		return nil, err
	}

	if credentials, err := options.GetCustomStringD("credentials", ""); credentials != "" {
		if credentials == "noauth" {
			clientOptions = append(clientOptions, option.WithoutAuthentication())
		} else {
			clientOptions = append(clientOptions, option.WithCredentialsJSON([]byte(credentials)))
		}
	} else if err != nil {
		return nil, err
	}

	client, err := storage.NewClient(
		context.TODO(),
		clientOptions...,
	)
	if err != nil {
		return nil, err
	}
	fs := GCSFilesystem{
		FilesystemOptions: options,
		client:            client,
		bucket:            client.Bucket(options.Host),
	}
	return &fs, nil
}

func (fs *GCSFilesystem) Write(filepath string, size int) (io.WriteCloser, error) {
	filepath = strings.TrimPrefix(filepath, "/")
	obj := fs.bucket.Object(filepath)
	return obj.NewWriter(context.TODO()), nil
}

// Read returns an opened File
func (fs *GCSFilesystem) Read(filepath string) (io.ReadCloser, error) {
	filepath = strings.TrimPrefix(filepath, "/")
	obj := fs.bucket.Object(filepath)
	r, err := obj.NewReader(context.TODO())
	if err != nil {
		if errors.Is(err, storage.ErrObjectNotExist) {
			err = os.ErrNotExist
		}
		return nil, err
	}
	return r, nil
}

// Delete implements the Filesystem interface
func (fs *GCSFilesystem) Delete(filepath string) error {
	filepath = strings.TrimPrefix(filepath, "/")
	obj := fs.bucket.Object(filepath)
	err := obj.Delete(context.TODO())
	if errors.Is(err, storage.ErrObjectNotExist) {
		return os.ErrNotExist
	}
	return err
}

// Rename renames a file
func (fs *GCSFilesystem) Rename(oldpath, newpath string) error {
	filepath1 := strings.TrimPrefix(oldpath, "/")
	filepath2 := strings.TrimPrefix(newpath, "/")
	obj1 := fs.bucket.Object(filepath1)
	obj2 := fs.bucket.Object(filepath2)
	cp := obj2.CopierFrom(obj1)
	_, err := cp.Run(context.TODO())
	var gerr *googleapi.Error
	if errors.As(err, &gerr) {
		if gerr.Code == 404 {
			err = os.ErrNotExist
		}
	}
	if err != nil {
		return err
	}
	return fs.Delete(oldpath)
}

// Observe implements the Filesystem interface
func (fs *GCSFilesystem) Observe(
	ctx context.Context, pattern string, paths chan<- string, log xbus.Logger,
) error {
	patterndir := path.Dir(path.Join(fs.Path, pattern))

	fo := NewFileObserver(ctx, pattern, fs.Path, paths)

	doList := func() ([]string, error) {
		var list []string
		it := fs.bucket.Objects(ctx, &storage.Query{
			Prefix: patterndir,
		})
		for {
			attrs, err := it.Next()
			if err == iterator.Done {
				break
			}
			if err != nil {
				return nil, err
			}
			list = append(list, attrs.Name)
		}
		return list, nil
	}

	fileList, err := doList()
	if err != nil {
		return err
	}
	select {
	case paths <- "":
	case <-ctx.Done():
		return ctx.Err()
	}

	if err := fo.AddFiles(fileList...); err != nil {
		log.Err(err).Msg("ftp: Error filtering file list")
	}
	var errcount int
	for {
		select {
		case <-time.After(fs.PollingInterval):
		case <-ctx.Done():
			return ctx.Err()
		}
		if fileList, err := doList(); err != nil {
			errcount++
		} else {
			if err := fo.AddFiles(fileList...); err != nil {
				log.Err(err).Msg("ftp: Error filtering file list")
				errcount++
			} else {
				errcount = 0
			}
		}
	}
}

// WaitForDeletion implements the Filesystem interface
func (fs *GCSFilesystem) WaitForDeletion(ctx context.Context, filepath string) error {
	filepath = strings.TrimPrefix(filepath, "/")
	for {
		it := fs.bucket.Objects(ctx, &storage.Query{
			Prefix: filepath,
		})
		var exists bool
		for {
			attrs, err := it.Next()
			if err == iterator.Done {
				break
			}
			if err != nil {
				return err
			}
			if attrs.Name == filepath {
				exists = true
			}
		}
		if !exists {
			return nil
		}
		select {
		case <-ctx.Done():
			return ctx.Err()
		case <-time.After(fs.PollingInterval):
		}
	}
}

func init() {
	MustRegisterFilesystem("gcs", gcsFilesystemFactory)
}
