package xbus2xbus

import (
	"context"
	"errors"
	"fmt"
	"sync"
	"time"

	nrpc "github.com/nats-rpc/nrpc"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	xbus_legacy "xbus.io/go-xbus/v4/legacy"
)

// Consumer emits incoming envelope in another bus.
type Consumer struct {
	log              xbus.Logger
	clientConfigFile string
	emitterName      string
	copyEnvelopeID   bool

	client    *xbus.Client
	emitterID api.UUID
}

// NewConsumer creates a Consumer.
func NewConsumer(actor *xbus.Actor, settings xbus.Settings) (xbus.ActorService, error) { //nolint:ireturn
	return &Consumer{
		log:              actor.Log,
		clientConfigFile: settings.MustString("target-client-config"),
		emitterName:      settings.MustString("target-emitter"),
		copyEnvelopeID:   settings.MustBoolD("copy-envelope-id", true),
	}, nil
}

// Startup connects to the downside xbus.
func (c *Consumer) Startup() error {
	options, err := xbus.COptionsFromFile(c.clientConfigFile, api.Account_ACTOR, c.log.Logger)
	if err != nil {
		return fmt.Errorf("xbus2xbus: %w", err)
	}

	options.ForceStart = true
	client := xbus.NewClient(options, c.log.Logger, nil)

	if err := client.Startup(false); err != nil {
		return fmt.Errorf("xbus2xbus: %w", err)
	}

	c.client = client
	if err := c.loadEmitter(); err != nil {
		c.client = nil

		return fmt.Errorf("xbus2xbus: %w", err)
	}

	return nil
}

// Shutdown disconnects from the downside xbus.
func (c *Consumer) Shutdown() error {
	c.client.Shutdown()

	return nil
}

var ErrNoSuchEmitter = errors.New("no such emitter")

func (c *Consumer) loadEmitter() error {
	if c.client.GetConn() != nil && c.client.GetConn().IsConnected() {
		if emitter := c.client.GetActorByName(c.emitterName); emitter != nil {
			c.emitterID = emitter.ID
		} else {
			c.client.Shutdown()

			return fmt.Errorf("%w: %s", ErrNoSuchEmitter, c.emitterName)
		}
	}

	return nil
}

type eventItem struct {
	eventType string
	item      api.Item
}

// Process forwards the incoming envelope to the downside xbus, or returns
// a SERVERTOOBUSY error.
func (c *Consumer) Process(apc *xbus.ActorProcessingContext) error { //nolint:funlen,cyclop
	ctx, cancel := context.WithCancel(apc.Context)
	defer cancel()

	if !c.client.GetConn().IsConnected() {
		return &nrpc.Error{ //nolint: exhaustivestruct
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: "[xbus2xbus] Not connected to downside xbus.",
		}
	}

	if c.emitterID.IsZero() {
		if err := c.loadEmitter(); err != nil {
			c.client = nil

			return err
		}
	}

	if c.emitterID.IsZero() {
		return &nrpc.Error{ //nolint:exhaustivestruct
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: "[xbus2xbus] Cannot get downside emitter ID.",
		}
	}

	reader, err := apc.ReadEnvelope("default")
	if err != nil {
		return fmt.Errorf("xbus2xbus: %w", err)
	}

	events := reader.Events()

	var eventIDs = make([]api.UUID, len(events))

	for i, event := range events {
		eventIDs[i] = event.GetIDAsUUID()
	}

	envID := reader.ID
	if !c.copyEnvelopeID {
		envID, err = api.UUIDV4()
		if err != nil {
			return fmt.Errorf("xbus2xbus: %w", err)
		}
	}

	output := xbus_legacy.NewLegacyEnvelopeSenderWithIDs(
		envID, reader.EventTypes(), eventIDs, api.EnvelopeMaxSize,
		func(env *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
			ack, err := c.client.SendEnvelopeFragment(c.emitterID, nil, "default", env)

			return ack, fmt.Errorf("xbus2xbus: %w", err)
		},
	)

	var wg sync.WaitGroup

	items := make(chan eventItem)

	for _, eventType := range reader.EventTypes() {
		itemChan := reader.GetItemChan(eventType)

		wg.Add(1)

		go func(eventType string, c <-chan api.Item) {
			defer wg.Done()

			for item := range c {
				select {
				case items <- eventItem{eventType, item}:
				case <-ctx.Done():
				}
			}
		}(eventType, itemChan)
	}

	go func() {
		wg.Wait()
		close(items)
	}()

	timeout := time.AfterFunc(time.Minute, func() {
		cancel()
	})
	for item := range items {
		timeout.Reset(time.Minute)

		if err := output.AddItems(item.eventType, item.item); err != nil {
			return fmt.Errorf("xbus2xbus: %w", err)
		}
	}

	if ctx.Err() != nil {
		reader.Cancel()
	}

	if ctx.Err() != nil || reader.Status() != api.EnvelopeAck_ACCEPTED {
		// TODO cancel emission output. //nolint:godox
		return fmt.Errorf("envelope forward to downstream failed") //nolint:goerr113
	}

	if err := output.Close(); err != nil {
		return fmt.Errorf("xbus2xbus: %w", err)
	}

	return nil
}

func init() {
	xbus.RegisterActorService("xbus2xbus-consumer", NewConsumer)
}
